//
// colortransf v. 1.0, (c)2014 Roman Starosolski, rstarosolski@polsl.pl
// 
// Apply color space transformation to RGB image obtaining transformed components, or recreate the RGB image from components using inverse transformation.
// Color space transformations are described in: R. Starosolski, New simple and efficient color space transformations for lossless image compression, Journal of Visual Communication and Image Representation, 2014, Vol. 25(5), pp. 1056-63, http://dx.doi.org/10.1016/j.jvcir.2014.03.003.
// Image formats: PPM P6 (for RGB image), PGM P5 (for components), PPM plane order RGB assumed, untransformed and transformed intensities assumed unsigned.
//
// This software is intended for research purposes only; it is provided "as is"; author makes no warranty of any kind, either express or implied, with respect to this software.
// This work was partially supported by POIG.02.03.01-24-099/13 grant: GCONiI - Upper-Silesian Center for Scientific Computations.
// 
// Commandline syntax information, supported transformations, examples - see synmsg below
//

const char *synmsg =
"\ncolortransf v. 1.0, (c)2014 Roman Starosolski, rstarosolski@polsl.pl "
"\n	"
"\nApply color space transformation to RGB image obtaining transformed components, or recreate the RGB image from components using inverse transformation."
"\nColor space transformations are described in: R. Starosolski, New simple and efficient color space transformations for lossless image compression, Journal of Visual Communication and Image Representation, 2014, Vol. 25(5), pp. 1056-63, http://dx.doi.org/10.1016/j.jvcir.2014.03.003."
"\nImage formats: PPM P6 (for RGB image), PGM P5 (for components), PPM plane order RGB assumed, untransformed and transformed intensities assumed unsigned."
"\n	"
"\nThis software is intended for research purposes only; it is provided \"as is\"; author makes no warranty of any kind, either express or implied, with respect to this software."
"\nThis work was partially supported by POIG.02.03.01-24-099/13 grant: GCONiI - Upper-Silesian Center for Scientific Computations."
"\n"
/*
\n1234567890123456789012345678901234567890123456789012345678901234567890123456789
*/
"\nsyntax: colortransf [switches] command"
"\n"
"\n commands: "
"\n  S infilergb.ppm redT.pgm greenT.pgm blueT.pgm  "
"\n    // split image into components, apply (optional) transformation"
"\n  U redT.pgm greenT.pgm blueT.pgm outfilergb.ppm "
"\n    // unsplit"
"\n  TEST_TRANSF range  "
"\n    // test Forward & Inverse rebersible transformation for all "
"\n    // possible combinations of component values in given range "
"\n    // checking dynamic ranges of components and reversibility"
"\n  "
"\n switches: "
"\n  [-TRANSF nr]  "
"\n    // apply transformation number nr (numbers listed below), for a detailed"
"\n    // description of transformations and components' formulas see the above"
"\n    // mentioned paper"
"\n    //"
"\n    // reversible transformations with inverses (* - 1-bit expansion):"
"\n    //  0  - RGB  (no transformation, default), "
"\n    //       components outputted in a following order: R G B          "
"\n    //  1  - RCT (from JPEG2000 standard),          "
"\n    //       components outputted in a following order: Cv* Y Cu*      "
"\n    // -1  - inverse RCT"
"\n    //  2  - YCoCg-R (from JPEG-XR standard),       "
"\n    //       components outputted in a following order: Co* Y Cg*      "
"\n    // -2  - inverse YCoCg-R"
"\n    //  3  - RDgDb,                                 "
"\n    //       components outputted in a following order: R Dg* Db*      "
"\n    // -3  - inverse RDgDb"
"\n    //  4  - A2,                                    "
"\n    //       components outputted in a following order: VA2* YA2 UA2*  "
"\n    // -4  - inverse A2"
"\n    //  5  - LDgEb,                                 "
"\n    //       components outputted in a following order: L Dg* Eb*      "
"\n    // -5  - inverse LDgEb"
"\n    //  6  - LDgDb,                                 "
"\n    //       components outputted in a following order: L Dg* Db*      "
"\n    // -6  - inverse LDgDb"
"\n    //  7  - mRCT (from extended JPEG-LS standard), "
"\n    //       components outputted in a following order: mCv mY mCu     "
"\n    // -7  - inverse mRCT"
"\n    //  8  - mRDgDb,                                "
"\n    //       components outputted in a following order: R mDg mDb      "
"\n    // -8  - inverse mRDgDb"
"\n    //  9  - mA2,                                   "
"\n    //       components outputted in a following order: mVA2 mYA2 mUA2 "
"\n    // -9  - inverse mA2"
"\n    //  10 - mLDgEb,                                "
"\n    //       components outputted in a following order: mL mDg mEb     "
"\n    // -10 - inverse mLDgEb"
"\n    //  11 - mLDgDb,                                "
"\n    //       components outputted in a following order: mL mDg mDb     "
"\n    // -11 - inverse mLDgDb"
"\n    //"
"\n    // for testing components' formulas (no inverse, * - 1-bit expansion):"
"\n    //  20 - ADDAVG, components outputted in a following order:   "
"\n    //                           ((2r+g+b)/4) ((2g+b+r)/4) ((2b+r+g)/4)"
"\n    //  21 - SUB, components outputted in a following order:   "
"\n    //                           (r-g)* (g-b)* (b-r)*"
"\n    //  22 - INVSUB, components outputted in a following order:   "
"\n    //                           (g-r)* (b-g)* (r-b)*"
"\n    //  23 - ADDHALF, components outputted in a following order:   "
"\n    //                           (r+g)/2 (g+b)/2 (b+r)/2"
"\n    //  24 - SUBAVG, components outputted in a following order:   "
"\n    //                           ((2r-g-b)/2)* ((2g-b-r)/2)* ((2b-r-g)/2*)"
"\n  [-O_RAW] "
"\n    // output as RAW (skip PGM/PPM header)"
"\n  [-NO_INV_CHECK] "
"\n    // skip checking inverse transformation (or reversibility), default is to"
"\n    // check for each pixel before outputting it (S and U commands); for the "
"\n    // TEST_TRANSF command check ranges only (skip reversibility check)"
"\n"
"\nexamples: "
"\n  Split lena.ppm RGB color image into red green and blue components, saved as"
"\n  grayscale images R.ppm G.ppm B.ppm:"
"\n      colortransf S lena.ppm R.pgm G.pgm B.pgm"
"\n  Transform lena.ppm using YCoCg-R transformation, save resulting components"
"\n  as Co.ppm Y.ppm Cg.ppm:"
"\n      colortransf -TRANSF 2 S lena.ppm Co.pgm Y.pgm Cg.pgm  "
"\n  Apply inverse YCoCg-R transformation on Co.pgm Y.pgm Cg.pgm components, save"
"\n  resulting image as lena2.ppm:"
"\n      colortransf -TRANSF -2 U Co.ppm Y.pgm Cg.pgm lena.pgm "
"\n  Convert lena.ppm RGB color image to BRG space and apply mRCT transformation"
"\n  on image with permuted components:"
"\n      colortransf S lena.ppm R.pgm G.pgm B.pgm "
"\n      colortransf U B.pgm R.pgm G.pgm lenaBRG.ppm "
"\n      colortransf -TRANSF 7 S lenaBRG.ppm mCvBRG.pgm mYBRG.pgm mCuBRG.pgm  "
"\n  Generate components being primary color differences (irreversible "
"\n  operation):"
"\n      colortransf -TRANSF 21 -NO_INV_CHECK S lena.ppm R-G.pgm G-B.pgm B-R.pgm"
"\n  Is mLDgEb really reversible for all possible combinations of component "
"\n  values in 0..255 range?"
"\n      colortransf -TRANSF 10 TEST_TRANSF 256"
"\n";


// disable fscanf and fopen compiler warnings
#define _CRT_SECURE_NO_WARNINGS

#include <cassert>
#include <cstdlib>
#include <cstdio>  
#include <cstring>  
#include <cctype>  



// globals ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------


int O_RAW;		    // save as RAW (skip PPM/PGM header), flag set in main(), see synmsg

int NO_INV_CHECK;	// skip checking reversibility of transformation for processed pixels in split() and unsplit(), 0 - do not skip, flag set in main(), see synmsg

int TRANSF_TYPE;	// transformation, set in main(), see synmsg

int RGB_range;		// size of the allowable range of untransformed RGB components (untransformed pixel components should be in 0 .. RGB_range-1)
                    // (as stored in .PPM header), for indiv. transformed components the range may be different, set by *_prepare


// messages ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------


void  DispSyntax()
{
	printf(synmsg);
}


// exit program and report an error 
void  ErrExit(const char * ErrMsg)
{
	fflush(stdout);
	fputs("\ncolortransf ERROR: ", stderr);
	fputs(ErrMsg, stderr);
	fputs("\n", stderr);
	fflush(stderr);
	exit(-1);
}


// helpers ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------


// bppmask[i] contains i ones as lsb-s
static const unsigned long int bppmask[33] = {
	0x00000000, /* [0] */
	0x00000001, 0x00000003, 0x00000007, 0x0000000f,
	0x0000001f, 0x0000003f, 0x0000007f, 0x000000ff,
	0x000001ff, 0x000003ff, 0x000007ff, 0x00000fff,
	0x00001fff, 0x00003fff, 0x00007fff, 0x0000ffff,
	0x0001ffff, 0x0003ffff, 0x0007ffff, 0x000fffff,
	0x001fffff, 0x003fffff, 0x007fffff, 0x00ffffff,
	0x01ffffff, 0x03ffffff, 0x07ffffff, 0x0fffffff,
	0x1fffffff, 0x3fffffff, 0x7fffffff, 0xffffffff /* [32] */
};


// ceil(log_2(val)) 
int ceil_log_2(int val)
{
	int result;

	assert(val>0);

	if (val == 1)
		return 0;

	result = 1;
	val -= 1;
	while (val >>= 1)
		result++;

	return result;
}


// x modulo range, regular modulo, it also works for range being not an integer power of 2
int modulo(int x, int range)
{
	assert(range>0);

	if (x >= 0)
		return (x % range);
	else
		return (range + (x % range));
}


// x symmetrical-modulo 2*halfrange, symmetrical modulo, result in the range -halfrange .. halfrange-1
int symmodulo(int x, int halfrange)
{
	return modulo(x + halfrange, 2 * halfrange) - halfrange;
}


// P?M file manipulation routines ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------


// image properties as in P?M header
struct imginfo
{
	char	ImgImageMode;
	int		ImgImageWidth, ImgImageHeight, ImgLevelMax;
};


// compare 2 imginfo-s
int imginfo_equal(imginfo * pi1, imginfo * pi2)
{
	return (pi1->ImgImageMode == pi2->ImgImageMode) && (pi1->ImgImageHeight == pi2->ImgImageHeight)
		&& (pi1->ImgImageWidth == pi2->ImgImageWidth) && (pi1->ImgLevelMax == pi2->ImgLevelMax);
}


//skip white spaces and comments in P?M header, return 0-ok, 1-error
//in f leave the first byte different than white space or comment
int __eatPgmSpacesAndComments(FILE * f)
{
	int i = -1;

	while (isspace(i = fgetc(f)));
	if (feof(f))
		return 1;
	while (i == '#')
	{
		while (fgetc(f) != 0x0a)
			if (feof(f))
				return 1;

		while (isspace(i = fgetc(f)));

		if (feof(f))
			return 1;
	}
	ungetc(i, f);

	return 0;
}


//read the P?M header
void  ReadPgmHeader(FILE * f, imginfo * pii)
{
	if (fscanf(f, "P%c", &(pii->ImgImageMode)) != 1)
		ErrExit("Bad P?M header 1");
	if (__eatPgmSpacesAndComments(f))
		ErrExit("Bad P?M header 2");

	if (fscanf(f, "%d %d", &(pii->ImgImageWidth), &(pii->ImgImageHeight)) != 2)
		ErrExit("Bad P?M header 3");
	if (__eatPgmSpacesAndComments(f))
		ErrExit("Bad P?M header 4");

	if (fscanf(f, "%d", &(pii->ImgLevelMax)) != 1)
		ErrExit("Bad P?M header 5");

	if (!(pii->ImgImageWidth>0 && pii->ImgImageHeight>0 && pii->ImgLevelMax>0 && pii->ImgLevelMax <= 0xFFFF))
		ErrExit("Bad P?M header 6");

	while (fgetc(f) != 0x0a)
		if (feof(f))
			ErrExit("Bad P?M header 7");
}


//write the P?M header
void  WritePgmHeader(FILE * f, imginfo * pii)
{
	if (EOF == fprintf(f, "P%c%c%d %d%c%d%c", pii->ImgImageMode, 0x0A, pii->ImgImageWidth,
		pii->ImgImageHeight, 0x0A, pii->ImgLevelMax, 0x0A))
		ErrExit("Error outputting P?M image header");
}


//write single pixel component or grayscale pixel, of bpp bit-depth
void  write_val_bpp(FILE *f, const int val, const int bpp)
{
	assert(!(val&~bppmask[bpp]));
	assert(bpp>0 && bpp <= 16);

	if (bpp>8)
		fputc((unsigned char)(val >> 8), f);
	if (fputc((unsigned char)val, f) == EOF) ErrExit("Error writting val");
}


//read single pixel component or grayscale pixel, of bpp bit-depth
int read_val_bpp(FILE *f, const int bpp)
{
	assert(bpp>0 && bpp <= 16);

	int hi = 0, lo, val;
	if (bpp>8)
		hi = (unsigned char)fgetc(f);
	lo = fgetc(f);
	if (lo == EOF) ErrExit("Error reading val");
	lo = (unsigned char)lo;

	val = hi << 8 | lo;

	if (val&~bppmask[bpp]) ErrExit("Read val out of bpp");

	return val;
}


// transformations ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------


/*
There are 4 functions for most transformations:
(exceptions:
-NOP transformation (which is just copying components) has only forward version, that may be used as inverse if needed
-transformations numbered from 20 to 24 have only forward version, for producing transformed components to be analyzed individually)

- void  FTRANSF_TransfName(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
do forward TransfName transformation for RGB pixel components r, g and b, store resulting transformed components in Rp, Gp and Bp

- void  ITRANSF_TransfName(const int Rp, const int Gp, const int Bp, int &r, int &g, int &b)
do inverse TransfName transformation for components Rp, Gp and Bp, store resulting untransformed RGB pixel components in r, g and b

- void  FTRANSF_TransfName_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
based on bit-depths of untransformed components r_bpp, g_bpp, b_bpp (in bpp, should be equal) store depths of transformed components in &outRbpp, &outGbpp, &outBbpp
set global variable RGB_range - size of the allowable range of untransformed components (untransformed pixel components should be in 0 .. RGB_range-1)

- void  ITRANSF_TransfName_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
based on bit-depths of transformed components r_bpp, g_bpp, b_bpp (in bpp) store depths of untransformed components in &outRbpp, &outGbpp, &outBbpp (will be equal)
set global variable RGB_range - size of the allowable range of untransformed components (untransformed pixel components should be in 0 .. RGB_range-1)

All input and output components are unsigned, declared as int for the purpose of intenal calculations
If their nominal range falls below 0, a correction is applied (usually +RGB_range/2 for luma, +RGB_range for chroma ) to make them >=0 before outputting (FTRANSF_*),
or after inputting (in corresponding ITRANSF_* - respectively -RGB_range/2 and +RGB_range).

*/


// RGB  (no transformation), components outputted in a following order: R G B (as Rp Gp Bp respectively) -------------------------------------------
void  TRANSF_NOP(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	Rp = r;
	Gp = g;
	Bp = b;
}


void  TRANSF_NOP_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}


// mRCT (from extended Jpeg-LS standard), components outputted: mCv mY mCu (as Rp Gp Bp respectively) -------------------------------------------
void  FTRANSF_mRCT(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	int rx, bx;

	rx = symmodulo(r - g, RGB_range / 2);
	bx = symmodulo(b - g, RGB_range / 2);
	Gp = modulo(g + ((rx + bx) >> 2), RGB_range);

	Rp = rx + RGB_range / 2;
	Bp = bx + RGB_range / 2;
}


// Inverse Transformation 
void  ITRANSF_mRCT(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	int rx, bx;

	rx = r - RGB_range / 2;
	bx = b - RGB_range / 2;

	Gp = modulo(g - ((rx + bx) >> 2), RGB_range);

	Rp = modulo(rx + Gp, RGB_range);
	Bp = modulo(bx + Gp, RGB_range);
}


void  FTRANSF_mRCT_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}


void  ITRANSF_mRCT_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}


// RCT (from JPEG2000 standard), components outputted: Cv Y Cu (as Rp Gp Bp respectively) -------------------------------------------
void  FTRANSF_RCT(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	int rx, bx;

	rx = r - g;
	bx = b - g;

	Gp = g + ((rx + bx) >> 2);
	Rp = rx + RGB_range;
	Bp = bx + RGB_range;
}


void  ITRANSF_RCT(const int Rp, const int Gp, const int Bp, int &r, int &g, int &b)
{
	int rx, bx;

	bx = Bp - RGB_range;
	rx = Rp - RGB_range;

	g = Gp - ((rx + bx) >> 2);
	b = bx + g;
	r = rx + g;
}


void  FTRANSF_RCT_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp + 1;
	outGbpp = g_bpp;
	outBbpp = b_bpp + 1;

	RGB_range = 0x01 << r_bpp;
}


void  ITRANSF_RCT_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp + 1) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = g_bpp;
	outGbpp = g_bpp;
	outBbpp = g_bpp;

	RGB_range = 0x01 << g_bpp;
}


// YCoCg-R (from JPEG-XR standard), components outputted: Co Y Cg -------------------------------------------
void  FTRANSF_YCoCg_R(const int r, const int g, const int b, int &Co, int &Y, int &Cg)
{
	int t, Co0, Cg0;

	Co0 = r - b;
	t = b + (Co0 >> 1);
	Cg0 = g - t;
	Y = t + (Cg0 >> 1);
	Co = Co0 + RGB_range;
	Cg = Cg0 + RGB_range;
}


// Inverse Reversible Colour Transform 
void  ITRANSF_YCoCg_R(const int Co, const int Y, const int Cg, int &r, int &g, int &b)
{
	int t, Co0, Cg0;

	Co0 = Co - RGB_range;
	Cg0 = Cg - RGB_range;
	t = Y - (Cg0 >> 1);
	g = Cg0 + t;
	b = t - (Co0 >> 1);
	r = Co0 + b;
}


void  FTRANSF_YCoCg_R_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp + 1;
	outGbpp = g_bpp;
	outBbpp = b_bpp + 1;

	RGB_range = 0x01 << r_bpp;
}


void  ITRANSF_YCoCg_R_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp + 1) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = g_bpp;
	outGbpp = g_bpp;
	outBbpp = g_bpp;

	RGB_range = 0x01 << g_bpp;
}


// RDgDb, components outputted: R Dg Db (as Rp Gp Bp respectively) -------------------------------------------
void  FTRANSF_RDGDB(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	Rp = r;
	Gp = r - g + RGB_range;
	Bp = g - b + RGB_range;
}


void  ITRANSF_RDGDB(const int Rp, const int Gp, const int Bp, int &r, int &g, int &b)
{
	r = Rp;
	g = r - Gp + RGB_range;
	b = g - Bp + RGB_range;
}


void  FTRANSF_RDGDB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp + 1;
	outBbpp = b_bpp + 1;

	RGB_range = 0x01 << r_bpp;
}

void  ITRANSF_RDGDB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp - 1) || (r_bpp != b_bpp - 1))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp - 1;
	outBbpp = b_bpp - 1;

	RGB_range = 0x01 << r_bpp;
}


// mRDgDb, components outputted: R mDg mDb (as Rp Gp Bp respectively) -------------------------------------------
void  FTRANSF_mRDGDB(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	Rp = r;
	Gp = symmodulo(r - g, RGB_range / 2) + RGB_range / 2;
	Bp = symmodulo(g - b, RGB_range / 2) + RGB_range / 2;
}


void  ITRANSF_mRDGDB(const int Rp, const int Gp, const int Bp, int &r, int &g, int &b)
{
	r = Rp;
	g = modulo(r - Gp + (RGB_range / 2), RGB_range);
	b = modulo(g - Bp + (RGB_range / 2), RGB_range);
}


void  FTRANSF_mRDGDB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}

void  ITRANSF_mRDGDB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}


// A2, components outputted: VA2 YA2 UA2 (as Rp Gp Bp respectively) -------------------------------------------
void  FTRANSF_A2(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	Rp = r - g + RGB_range;
	Gp = g;
	Bp = b - g + RGB_range;
}


void  ITRANSF_A2(const int Rp, const int Gp, const int Bp, int &r, int &g, int &b)
{
	g = Gp;
	r = Rp + g - RGB_range;
	b = Bp + g - RGB_range;
}


void  FTRANSF_A2_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp + 1;
	outGbpp = g_bpp;
	outBbpp = b_bpp + 1;

	RGB_range = 0x01 << g_bpp;
}


void  ITRANSF_A2_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((g_bpp != r_bpp - 1) || (g_bpp != b_bpp - 1))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp - 1;
	outGbpp = g_bpp;
	outBbpp = b_bpp - 1;

	RGB_range = 0x01 << g_bpp;
}


// mA2, components outputted: mVA2 mYA2 mUA2 (as Rp Gp Bp respectively) -------------------------------------------
void  FTRANSF_mA2(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	Rp = symmodulo(r - g, RGB_range / 2) + RGB_range / 2;
	Gp = g;
	Bp = symmodulo(b - g, RGB_range / 2) + RGB_range / 2;
}


void  ITRANSF_mA2(const int Rp, const int Gp, const int Bp, int &r, int &g, int &b)
{
	g = Gp;
	r = modulo(Rp + g + (RGB_range / 2), RGB_range);
	b = modulo(Bp + g + (RGB_range / 2), RGB_range);
}


void  FTRANSF_mA2_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}


void  ITRANSF_mA2_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}


// LDgEb, components outputted: L Dg Eb (as Rp Gp Bp respectively) -------------------------------------------
void  FTRANSF_LDGEB(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	int T;

	T = r - g;
	Rp = r - (T >> 1);
	Gp = T + RGB_range;
	Bp = b - Rp + RGB_range;
}


void  ITRANSF_LDGEB(const int Rp, const int Gp, const int Bp, int &r, int &g, int &b)
{
	int T;

	T = Gp - RGB_range;
	r = Rp + (T >> 1);
	g = r - T;
	b = Bp + Rp - RGB_range;
}


void  FTRANSF_LDGEB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp + 1;
	outBbpp = b_bpp + 1;

	RGB_range = 0x01 << r_bpp;
}

void  ITRANSF_LDGEB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp - 1) || (r_bpp != b_bpp - 1))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp - 1;
	outBbpp = b_bpp - 1;

	RGB_range = 0x01 << r_bpp;
}


// mLDgEb, components outputted: mL mDg mEb (as Rp Gp Bp respectively) -------------------------------------------
void  FTRANSF_mLDGEB(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	int T;

	T = symmodulo(r - g, RGB_range / 2);
	Rp = modulo(r - (T >> 1), RGB_range);
	Gp = T + (RGB_range / 2);
	Bp = symmodulo(b - Rp, RGB_range / 2) + RGB_range / 2;
}


void  ITRANSF_mLDGEB(const int Rp, const int Gp, const int Bp, int &r, int &g, int &b)
{
	int T;

	T = Gp - (RGB_range / 2);
	r = modulo(Rp + (T >> 1), RGB_range);
	g = modulo(r - T, RGB_range);
	b = modulo(Bp + Rp - (RGB_range / 2), RGB_range);
}


void  FTRANSF_mLDGEB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}

void  ITRANSF_mLDGEB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}


// LDgDb, components outputted: L Dg Db (as Rp Gp Bp respectively) -------------------------------------------
void  FTRANSF_LDGDB(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	int T;

	T = r - g;
	Rp = r - (T >> 1);
	Gp = T + RGB_range;
	Bp = g - b + RGB_range;
}


void  ITRANSF_LDGDB(const int Rp, const int Gp, const int Bp, int &r, int &g, int &b)
{
	int T;

	T = Gp - RGB_range;
	r = Rp + (T >> 1);
	g = r - T;
	b = g - Bp + RGB_range;
}


void  FTRANSF_LDGDB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp + 1;
	outBbpp = b_bpp + 1;

	RGB_range = 0x01 << r_bpp;
}

void  ITRANSF_LDGDB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp - 1) || (r_bpp != b_bpp - 1))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp - 1;
	outBbpp = b_bpp - 1;

	RGB_range = 0x01 << r_bpp;
}


// mLDgDb, components outputted: mL mDg mDb (as Rp Gp Bp respectively) -------------------------------------------
void  FTRANSF_mLDGDB(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	int T;

	T = symmodulo(r - g, RGB_range / 2);
	Rp = modulo(r - (T >> 1), RGB_range);
	Gp = T + (RGB_range / 2);
	Bp = symmodulo(g - b, RGB_range / 2) + RGB_range / 2;
}


void  ITRANSF_mLDGDB(const int Rp, const int Gp, const int Bp, int &r, int &g, int &b)
{
	int T;

	T = Gp - (RGB_range / 2);
	r = modulo(Rp + (T >> 1), RGB_range);
	g = modulo(r - T, RGB_range);
	b = modulo(g - Bp + (RGB_range / 2), RGB_range);
}


void  FTRANSF_mLDGDB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}


void  ITRANSF_mLDGDB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}


// ADDAVG,   components outputted: (2r+g+b)/4 (2g+b+r)/4 (2b+r+g)/4 (as Rp Gp Bp respectively), for testing, no inverse -------------------------------------------
void  FTRANSF_ADDAVG(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	Rp = (2 * r + g + b) >> 2;
	Gp = (2 * g + b + r) >> 2;
	Bp = (2 * b + r + g) >> 2;
}

void  FTRANSF_ADDAVG_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}


// SUB,      components outputted: (r-g) (g-b) (b-r) (as Rp Gp Bp respectively), for testing, no inverse -------------------------------------------
void  FTRANSF_SUB(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	Rp = r - g + RGB_range;
	Gp = g - b + RGB_range;
	Bp = b - r + RGB_range;
}


void  FTRANSF_SUB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp + 1;
	outGbpp = g_bpp + 1;
	outBbpp = b_bpp + 1;

	RGB_range = 0x01 << r_bpp;
}


// INVSUB,   components outputted: (g-r) (b-g) (r-b) (as Rp Gp Bp respectively), for testing, no inverse -------------------------------------------
void  FTRANSF_INVSUB(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	Rp = g - r + RGB_range;
	Gp = b - g + RGB_range;
	Bp = r - b + RGB_range;
}


void  FTRANSF_INVSUB_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp + 1;
	outGbpp = g_bpp + 1;
	outBbpp = b_bpp + 1;

	RGB_range = 0x01 << r_bpp;
}


// ADDHALF,  components outputted: (r+g)/2 (g+b)/2 (b+r)/2 (as Rp Gp Bp respectively), for testing, no inverse -------------------------------------------
void  FTRANSF_ADDHALF(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	Rp = (r + g) >> 1;
	Gp = (g + b) >> 1;
	Bp = (b + r) >> 1;
}


void  FTRANSF_ADDHALF_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp;
	outGbpp = g_bpp;
	outBbpp = b_bpp;

	RGB_range = 0x01 << r_bpp;
}


// SUBAVG,  components outputted: (2r-g-b)/2 (2g-b-r)/2 (2b-r-g)/2 (as Rp Gp Bp respectively), for testing, no inverse -------------------------------------------
void  FTRANSF_SUBAVG(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	Rp = ((2 * r - g - b) >> 1) + RGB_range;
	Gp = ((2 * g - b - r) >> 1) + RGB_range;
	Bp = ((2 * b - r - g) >> 1) + RGB_range;
}


void  FTRANSF_SUBAVG_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	if ((r_bpp != g_bpp) || (r_bpp != b_bpp))
		ErrExit("Input colorspace plane Range missmatch");

	outRbpp = r_bpp + 1;
	outGbpp = g_bpp + 1;
	outBbpp = b_bpp + 1;

	RGB_range = 0x01 << r_bpp;
}


// commandline commands ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------


// calls transformation detrmined by global TRANSF_TYPE
void  TRANSF(const int r, const int g, const int b, int &Rp, int &Gp, int &Bp)
{
	switch (TRANSF_TYPE)
	{
	case 0: return TRANSF_NOP(r, g, b, Rp, Gp, Bp);
	case 7: return FTRANSF_mRCT(r, g, b, Rp, Gp, Bp);
	case -7: return ITRANSF_mRCT(r, g, b, Rp, Gp, Bp);
	case 1: return FTRANSF_RCT(r, g, b, Rp, Gp, Bp);
	case -1: return ITRANSF_RCT(r, g, b, Rp, Gp, Bp);
	case 2: return FTRANSF_YCoCg_R(r, g, b, Rp, Gp, Bp);
	case -2: return ITRANSF_YCoCg_R(r, g, b, Rp, Gp, Bp);
	case 3: return FTRANSF_RDGDB(r, g, b, Rp, Gp, Bp);
	case -3: return ITRANSF_RDGDB(r, g, b, Rp, Gp, Bp);
	case 4: return FTRANSF_A2(r, g, b, Rp, Gp, Bp);
	case -4: return ITRANSF_A2(r, g, b, Rp, Gp, Bp);
	case 5: return FTRANSF_LDGEB(r, g, b, Rp, Gp, Bp);
	case -5: return ITRANSF_LDGEB(r, g, b, Rp, Gp, Bp);
	case 6: return FTRANSF_LDGDB(r, g, b, Rp, Gp, Bp);
	case -6: return ITRANSF_LDGDB(r, g, b, Rp, Gp, Bp);
	case 8: return FTRANSF_mRDGDB(r, g, b, Rp, Gp, Bp);
	case -8: return ITRANSF_mRDGDB(r, g, b, Rp, Gp, Bp);
	case 9: return FTRANSF_mA2(r, g, b, Rp, Gp, Bp);
	case -9: return ITRANSF_mA2(r, g, b, Rp, Gp, Bp);
	case 10: return FTRANSF_mLDGEB(r, g, b, Rp, Gp, Bp);
	case -10: return ITRANSF_mLDGEB(r, g, b, Rp, Gp, Bp);
	case 11: return FTRANSF_mLDGDB(r, g, b, Rp, Gp, Bp);
	case -11: return ITRANSF_mLDGDB(r, g, b, Rp, Gp, Bp);

	case 20: return FTRANSF_ADDAVG(r, g, b, Rp, Gp, Bp);
	case 21: return FTRANSF_SUB(r, g, b, Rp, Gp, Bp);
	case 22: return FTRANSF_INVSUB(r, g, b, Rp, Gp, Bp);
	case 23: return FTRANSF_ADDHALF(r, g, b, Rp, Gp, Bp);
	case 24: return FTRANSF_SUBAVG(r, g, b, Rp, Gp, Bp);

	default:
		ErrExit("In TRANSF() bad TRANSF_TYPE");
	}
}


// calls *_prepare function of transformation detrmined by global TRANSF_TYPE
// check if r_bpp, int g_bpp, b_bpp input colorspace dynamic ranges are compatible with the transformation
// report output dynamic range of components in outRbpp, outGbpp, outBbpp
void  TRANSF_prepare(const int r_bpp, const int g_bpp, const int b_bpp, int &outRbpp, int &outGbpp, int &outBbpp)
{
	switch (TRANSF_TYPE)
	{
	case 0: return TRANSF_NOP_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 7: return FTRANSF_mRCT_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case -7: return ITRANSF_mRCT_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 1: return FTRANSF_RCT_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case -1: return ITRANSF_RCT_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 2: return FTRANSF_YCoCg_R_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case -2: return ITRANSF_YCoCg_R_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 3: return FTRANSF_RDGDB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case -3: return ITRANSF_RDGDB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 4: return FTRANSF_A2_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case -4: return ITRANSF_A2_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 5: return FTRANSF_LDGEB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case -5: return ITRANSF_LDGEB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 6: return FTRANSF_LDGDB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case -6: return ITRANSF_LDGDB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 8: return FTRANSF_mRDGDB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case -8: return ITRANSF_mRDGDB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 9: return FTRANSF_mA2_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case -9: return ITRANSF_mA2_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 10: return FTRANSF_mLDGEB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case -10: return ITRANSF_mLDGEB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 11: return FTRANSF_mLDGDB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case -11: return ITRANSF_mLDGDB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);

	case 20: return FTRANSF_ADDAVG_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 21: return FTRANSF_SUB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 22: return FTRANSF_INVSUB_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 23: return FTRANSF_ADDHALF_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);
	case 24: return FTRANSF_SUBAVG_prepare(r_bpp, g_bpp, b_bpp, outRbpp, outGbpp, outBbpp);

	default:
		ErrExit("In TRANSF_prepare() bad TRANSF_TYPE");
	}
}


// commandline commands ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------


// command TEST_TRANSF: test if FTRANSF and ITRANSF compatible with each other and with transformation component depths for components in range 0 .. range-1
void  TEST_TRANSF(int range)
{
	int in_bpp, inRbpp, inGbpp, inBbpp, outRbpp, outGbpp, outBbpp;
	int fwd_range, inv_range;
	int fwd_RCT_type, inv_RCT_type;

	printf("\nwait . . .");

	in_bpp = ceil_log_2(range);
	inRbpp = inGbpp = inBbpp = in_bpp;

	TRANSF_prepare(inRbpp, inGbpp, inBbpp, outRbpp, outGbpp, outBbpp);
	fwd_range = RGB_range;
	fwd_RCT_type = TRANSF_TYPE;

	if (!NO_INV_CHECK)
	{
		inv_RCT_type = TRANSF_TYPE = -TRANSF_TYPE;
		TRANSF_prepare(outRbpp, outGbpp, outBbpp, inRbpp, inGbpp, inBbpp);
		inv_range = RGB_range;
		if ((inRbpp != in_bpp) || (inGbpp != in_bpp) || (inBbpp != in_bpp))
			ErrExit("ERROR! FTRANSF/ITRANSF Range missmatch");
	}

	for (int r = 0; r<range; r++)
		for (int g = 0; g<range; g++)
			for (int b = 0; b<range; b++)
			{
				int R, G, B, rr, gg, bb;

				TRANSF_TYPE = fwd_RCT_type;
				RGB_range = fwd_range;
				TRANSF(r, g, b, R, G, B);

				if ((R&~bppmask[outRbpp]) || (G&~bppmask[outGbpp]) || (B&~bppmask[outBbpp]))
				{
					printf("\n\nERROR! FTRANSF fail (violate out_bpp) for r=%d, g=%d b=%d\n", r, g, b);
					ErrExit("Error: TEST_TRANSF");
				}

				if (!NO_INV_CHECK)
				{
					TRANSF_TYPE = inv_RCT_type;
					RGB_range = inv_range;
					TRANSF(R, G, B, rr, gg, bb);

					if ((r != rr) || (g != gg) || (b != bb))
					{
						printf("\n\nERROR! FTRANSF/ITRANSF fail for r=%d, g=%d b=%d\n", r, g, b);
						ErrExit("Error: TEST_TRANSF");
					}
				}
			}

	TRANSF_TYPE = fwd_RCT_type;
	printf("\nFTRANSF/ITRANSF working for all combinations of rgb color values in range 0 .. %d\n", range - 1);
}


// command S, input and output files read from commandline reminder passed by argv
void  split(char **argv)
{
	FILE * infile, *out_r, *out_g, *out_b;
	struct imginfo infoin, infoout_r, infoout_g, infoout_b;
	int in_bpp, outRbpp, outGbpp, outBbpp;

	if (!*argv) ErrExit("Not enough cmdline parameters");
	infile = fopen(*(argv++), "rb");
	if (!infile) ErrExit("Cannot open input");
	ReadPgmHeader(infile, &infoin);
	if (infoin.ImgImageMode != '6') ErrExit("Bad input format - not PPM P6");

	in_bpp = ceil_log_2(infoin.ImgLevelMax + 1);
	TRANSF_prepare(in_bpp, in_bpp, in_bpp, outRbpp, outGbpp, outBbpp);

	infoout_r = infoin;
	infoout_r.ImgImageMode = '5';
	infoout_r.ImgLevelMax = bppmask[outRbpp];
	infoout_g = infoout_r;
	infoout_g.ImgLevelMax = bppmask[outGbpp];
	infoout_b = infoout_r;
	infoout_b.ImgLevelMax = bppmask[outBbpp];

	if (!*argv) ErrExit("Not enough cmdline parameters");
	out_r = fopen(*(argv++), "wb");
	if (!out_r) ErrExit("Cannot open output");
	if (!O_RAW)
		WritePgmHeader(out_r, &infoout_r);

	if (!*argv) ErrExit("Not enough cmdline parameters");
	out_g = fopen(*(argv++), "wb");
	if (!out_g) ErrExit("Cannot open output");
	if (!O_RAW)
		WritePgmHeader(out_g, &infoout_g);

	if (!*argv) ErrExit("Not enough cmdline parameters");
	out_b = fopen(*(argv++), "wb");
	if (!out_b) ErrExit("Cannot open output");
	if (!O_RAW)
		WritePgmHeader(out_b, &infoout_b);

	if (*argv) ErrExit("Too much cmdline parameters");

	for (int pixels = infoin.ImgImageHeight*infoin.ImgImageWidth; pixels; pixels--)
	{
		int r, g, b, R, G, B;

		r = read_val_bpp(infile, in_bpp);
		g = read_val_bpp(infile, in_bpp);
		b = read_val_bpp(infile, in_bpp);

		TRANSF(r, g, b, R, G, B);

		if (!NO_INV_CHECK)
		{//test reversibility
			int rr, gg, bb;
			TRANSF_TYPE = -TRANSF_TYPE;
			TRANSF(R, G, B, rr, gg, bb);
			if ((r != rr) || (g != gg) || (b != bb))
			{
				printf("\n\nERROR! FTRANSF/ITRANSF fail for r=%d, g=%d b=%d\n", r, g, b);
				ErrExit("Error: split");
			}
			TRANSF_TYPE = -TRANSF_TYPE;
		}

		write_val_bpp(out_r, R, outRbpp);
		write_val_bpp(out_g, G, outGbpp);
		write_val_bpp(out_b, B, outBbpp);
	}

	if (fgetc(infile) != EOF) ErrExit("Error infile too long");

	fclose(infile);
	fclose(out_r);
	fclose(out_g);
	fclose(out_b);
}


// command U, input and output files read from commandline reminder passed by argv
void  unsplit(char **argv) // U
{
	FILE * in_r, *in_g, *in_b, *outfile;
	struct imginfo infoin_r, infoin_g, infoin_b, infoout;
	int inRbpp, inGbpp, inBbpp, out_bpp, ob1, ob2;

	if (!*argv) ErrExit("Not enough cmdline parameters");
	in_r = fopen(*(argv++), "rb");
	if (!in_r) ErrExit("Cannot open input");
	ReadPgmHeader(in_r, &infoin_r);
	if (infoin_r.ImgImageMode != '5') ErrExit("Bad input format - not PPM P5");
	inRbpp = ceil_log_2(infoin_r.ImgLevelMax + 1);

	if (!*argv) ErrExit("Not enough cmdline parameters");
	in_g = fopen(*(argv++), "rb");
	if (!in_g) ErrExit("Cannot open input");
	ReadPgmHeader(in_g, &infoin_g);
	if (infoin_g.ImgImageMode != '5') ErrExit("Bad input format - not PPM P5");
	if ((infoin_r.ImgImageHeight != infoin_g.ImgImageHeight) || (infoin_r.ImgImageWidth != infoin_g.ImgImageWidth))
		ErrExit("Bad input format - incompatibile input files r&g");
	inGbpp = ceil_log_2(infoin_g.ImgLevelMax + 1);

	if (!*argv) ErrExit("Not enough cmdline parameters");
	in_b = fopen(*(argv++), "rb");
	if (!in_b) ErrExit("Cannot open input");
	ReadPgmHeader(in_b, &infoin_b);
	if (infoin_b.ImgImageMode != '5') ErrExit("Bad input format - not PPM P5");
	if ((infoin_r.ImgImageHeight != infoin_b.ImgImageHeight) || (infoin_r.ImgImageWidth != infoin_b.ImgImageWidth))
		ErrExit("Bad input format - incompatibile input files r&b");
	inBbpp = ceil_log_2(infoin_b.ImgLevelMax + 1);

	TRANSF_prepare(inRbpp, inGbpp, inBbpp, out_bpp, ob1, ob2);
	if ((out_bpp != ob1) || (out_bpp != ob2))
		ErrExit("(out_bpp!=ob1) || (out_bpp!=ob2)");

	infoout = infoin_r;
	infoout.ImgImageMode = '6';
	infoout.ImgLevelMax = bppmask[out_bpp];

	if (!*argv) ErrExit("Not enough cmdline parameters");
	outfile = fopen(*(argv++), "wb");
	if (!outfile) ErrExit("Cannot open output");
	if (!O_RAW)
		WritePgmHeader(outfile, &infoout);

	if (*argv) ErrExit("Too much cmdline parameters");

	for (int pixels = infoin_r.ImgImageHeight*infoin_r.ImgImageWidth; pixels; pixels--)
	{
		int r, g, b, R, G, B;

		r = read_val_bpp(in_r, inRbpp);
		g = read_val_bpp(in_g, inGbpp);
		b = read_val_bpp(in_b, inBbpp);

		TRANSF(r, g, b, R, G, B);

		if (!NO_INV_CHECK)
		{//test reversibility
			int rr, gg, bb;
			TRANSF_TYPE = -TRANSF_TYPE;
			TRANSF(R, G, B, rr, gg, bb);
			if ((r != rr) || (g != gg) || (b != bb))
			{
				printf("\n\nERROR! FTRANSF/ITRANSF fail for r=%d, g=%d b=%d\n", r, g, b);
				ErrExit("Error: unsplit");
			}
			TRANSF_TYPE = -TRANSF_TYPE;
		}

		write_val_bpp(outfile, R, out_bpp);
		write_val_bpp(outfile, G, out_bpp);
		write_val_bpp(outfile, B, out_bpp);
	}

	if (fgetc(in_r) != EOF) ErrExit("Error infile too long");
	if (fgetc(in_g) != EOF) ErrExit("Error infile too long");
	if (fgetc(in_b) != EOF) ErrExit("Error infile too long");

	fclose(in_r);
	fclose(in_g);
	fclose(in_b);
	fclose(outfile);
}


// main ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------


int main(int, char **argv)
{
	O_RAW = 0;
	TRANSF_TYPE = 0;
	NO_INV_CHECK = 0;

	argv++;
	if (!*argv)
	{
		DispSyntax();
		return 0;
	}

	// switch
	while (**argv == '-')
	{
		if (!strcmp(*argv, "-O_RAW"))
		{
			argv++;
			O_RAW = 1;
			printf("\nOutputting as RAW");
		}
		if (!strcmp(*argv, "-TRANSF"))
		{
			argv++;
			if (!*argv)
			{
				DispSyntax();
				return 0;
			}
			TRANSF_TYPE = atoi(*argv);
			argv++;
			printf("\nTRANSF_TYPE:%d", TRANSF_TYPE);
		}
		if (!strcmp(*argv, "-NO_INV_CHECK"))
		{
			argv++;
			NO_INV_CHECK = 1;
			printf("\nNO_INV_CHECK set, skiping reversibility check");
		}

	}

	// command
	if (!strcmp(*argv, "S"))
	{
		argv++;
		split(argv);
	}
	else if (!strcmp(*argv, "U"))
	{
		argv++;
		unsplit(argv);
	}
	else if (!strcmp(*argv, "TEST_TRANSF"))
	{
		int range;
		argv++;
		if (!*argv)
		{
			DispSyntax();
			return 0;
		}
		range = atoi(*argv);
		argv++;
		if (*argv) ErrExit("Too much cmdline parameters");

		TEST_TRANSF(range);
	}
	else
	{
		DispSyntax();
	}

	return 0;
}

